"""Context definition for exit condition programs."""
import datetime
from typing import Any, Dict, Iterable, Optional, Sequence

from routemaster.feeds import Feed
from routemaster.utils import get_path


class Context(object):
    """Execution context for exit condition programs."""

    def __init__(
        self,
        *,
        label: str,
        metadata: Dict[str, Any],
        now: datetime.datetime,
        feeds: Dict[str, Feed],
        accessed_variables: Iterable[str],
        current_history_entry: Optional[Any],
        feed_logging_context,
    ) -> None:
        """Create an execution context."""
        if now.tzinfo is None:
            raise ValueError(
                "Cannot evaluate exit conditions with naive datetimes",
            )

        self.now = now
        self.metadata = metadata
        self.feeds = feeds
        self.current_history_entry = current_history_entry

        self._pre_warm_feeds(label, accessed_variables, feed_logging_context)

    def lookup(self, path: Sequence[str]) -> Any:
        """Look up a path in the execution context."""
        location, *rest = path

        try:
            # Changing this mapping? Also change config validation in
            # `routemaster.config.loader._validate_context_lookups`
            return {
                'metadata': self._lookup_metadata,
                'feeds': self._lookup_feed_data,
                'history': self._lookup_history,
            }[location](rest)
        except (KeyError, ValueError):
            return None

    def _lookup_metadata(self, path: Sequence[str]) -> Any:
        return get_path(path, self.metadata)

    def _lookup_feed_data(self, path: Sequence[str]) -> Any:
        feed_name, *rest = path
        return self.feeds[feed_name].lookup(rest)

    def _lookup_history(self, path: Sequence[str]) -> Any:
        if self.current_history_entry is None:
            raise ValueError("Accessed uninitialised variable")

        variable_name, = path
        return {
            'entered_state': self.current_history_entry.created,
            'previous_state': self.current_history_entry.old_state,
        }[variable_name]

    def property_handler(self, property_name, value, **kwargs):
        """Handle a property in execution."""
        if property_name == ('passed',):
            epoch = kwargs['since']
            return (self.now - epoch).total_seconds() >= value
        if property_name == ('defined',):
            return value is not None
        if property_name == () and 'in' in kwargs:
            return value in kwargs['in']
        raise ValueError("Unknown property {name}".format(
            name='.'.join(property_name)),
        )

    def _pre_warm_feeds(
        self,
        label: str,
        accessed_variables: Iterable[str],
        logging_context,
    ):
        for accessed_variable in accessed_variables:
            parts = accessed_variable.split('.')

            if len(parts) < 2:
                continue

            if parts[0] != 'feeds':
                continue

            feed = self.feeds.get(parts[1])
            if feed is not None:
                with logging_context(feed.url) as log_response:
                    feed.prefetch(label, log_response)



"""Database model definition."""
import datetime
import functools
from typing import Any

import dateutil.tz
from sqlalchemy import DDL, Table
from sqlalchemy import Column as NullableColumn
from sqlalchemy import (
    String,
    Boolean,
    Integer,
    DateTime,
    MetaData,
    FetchedValue,
    ForeignKeyConstraint,
    func,
)
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import JSONB

metadata = MetaData()

Base: Any  # Workaround for https://github.com/python/mypy/issues/2477
Base = declarative_base(metadata=metadata)

Column = functools.partial(NullableColumn, nullable=False)

sync_label_updated_column = DDL(
    '''
    CREATE OR REPLACE FUNCTION sync_label_updated_column_fn()
        RETURNS TRIGGER AS
            $$
                BEGIN
                    NEW.updated = now();
                    RETURN NEW;
                END;
            $$
        LANGUAGE PLPGSQL;

    CREATE TRIGGER sync_label_updated_column
        BEFORE UPDATE ON labels
        FOR EACH ROW
        EXECUTE PROCEDURE sync_label_updated_column_fn();
    ''',
)


# ORM classes


class Label(Base):
    """A single label including context."""

    # Note: type annotations for this class are provided by a stubs file

    __table__ = Table(
        'labels',
        metadata,
        Column('name', String, primary_key=True),
        Column('state_machine', String, primary_key=True),
        Column('metadata', JSONB),
        Column('metadata_triggers_processed', Boolean, default=True),
        Column('deleted', Boolean, default=False),
        Column(
            'updated',
            DateTime(timezone=True),
            server_default=func.now(),
            server_onupdate=FetchedValue(),
        ),
        listeners=[
            ('after_create', sync_label_updated_column),
        ],
    )

    def __repr__(self):
        """Return a useful debug representation."""
        return (
            f"Label(state_machine={self.state_machine!r}, name={self.name!r})"
        )


class History(Base):
    """A single historical state transition of a label."""

    # Note: type annotations for this class are provided by a stubs file

    __table__ = Table(
        'history',
        metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),

        Column('label_name', String),
        Column('label_state_machine', String),
        ForeignKeyConstraint(
            ['label_name', 'label_state_machine'],
            ['labels.name', 'labels.state_machine'],
        ),

        Column(
            'created',
            DateTime(timezone=True),
            default=lambda: datetime.datetime.now(dateutil.tz.tzutc()),
        ),

        # `forced = True` represents a manual transition that may not be in
        # accordance with the state machine logic.
        Column('forced', Boolean, default=False),

        # Null indicates starting a state machine
        NullableColumn('old_state', String),

        # Null indicates being deleted from a state machine
        NullableColumn('new_state', String),
    )

    label = relationship(Label, backref='history')

    def __repr__(self):
        """Return a useful debug representation."""
        return (
            f"History(id={self.id!r}, "
            f"label_state_machine={self.label_state_machine!r}, "
            f"label_name={self.label_name!r})"
        )



"""Tokenization of exit condition programs."""

import re
import enum
import unicodedata
from typing import Any, Tuple, Iterable, NamedTuple

from routemaster.exit_conditions.exceptions import ParseError
from routemaster.exit_conditions.prepositions import Preposition

RE_DURATION = re.compile(r'^(?:(\d+)d)?(?:(\d+)h)?(?:(\d+)m)?(?:(\d+)s)?$')


@enum.unique
class RawTokenKind(enum.Enum):
    """Types of (pre-digestion) program token."""
    LEFT_PAREN = 'LEFT_PAREN'
    RIGHT_PAREN = 'RIGHT_PAREN'
    ATOM = 'ATOM'
    OPERATOR = 'OPERATOR'
    COMMENT = 'COMMENT'
    WHITESPACE = 'WHITESPACE'


@enum.unique
class TokenKind(enum.Enum):
    """Types of major program token."""

    LEFT_PAREN = '"("'
    RIGHT_PAREN = '")"'
    ATOM = 'atom'
    OPERATOR = 'operator'
    NUMBER = 'number'
    DURATION = 'duration'
    AND = '"and"'
    OR = '"or"'
    NOT = '"not"'
    LITERAL = 'constant'
    COPULA = 'is/has'
    PREPOSITION = 'preposition'


RAW_TOKEN_KIND_TO_TOKEN_KIND = {
    RawTokenKind.LEFT_PAREN: TokenKind.LEFT_PAREN,
    RawTokenKind.RIGHT_PAREN: TokenKind.RIGHT_PAREN,
    RawTokenKind.OPERATOR: TokenKind.OPERATOR,
}


class RawToken(NamedTuple):
    """A single raw (in-text, undigested) token."""

    kind: RawTokenKind
    value: str
    location: Tuple[int, int]


class Token(NamedTuple):
    """A single digested (usable) token."""

    kind: TokenKind
    value: Any
    location: Tuple[int, int]


LITERALS = {
    'true': (TokenKind.LITERAL, True),
    'false': (TokenKind.LITERAL, False),
    'null': (TokenKind.LITERAL, None),
    'and': (TokenKind.AND, None),
    'or': (TokenKind.OR, None),
    'not': (TokenKind.NOT, None),
    'is': (TokenKind.COPULA, 'is'),
    'has': (TokenKind.COPULA, 'has'),
    'was': (TokenKind.COPULA, 'was'),
}

STATE_MACHINE = {
    (None, 'Ps('): RawTokenKind.LEFT_PAREN,
    (None, 'Pe)'): RawTokenKind.RIGHT_PAREN,
    (None, 'Z'): RawTokenKind.WHITESPACE,
    (None, 'Cc'): RawTokenKind.WHITESPACE,
    (None, 'Po#'): RawTokenKind.COMMENT,
    (None, 'L'): RawTokenKind.ATOM,
    (None, 'Pd'): RawTokenKind.ATOM,
    (None, 'Pc'): RawTokenKind.ATOM,
    (None, 'N'): RawTokenKind.ATOM,
    (None, 'S'): RawTokenKind.OPERATOR,
    (None, 'Po'): RawTokenKind.OPERATOR,
    (RawTokenKind.WHITESPACE, 'Ps('): RawTokenKind.LEFT_PAREN,
    (RawTokenKind.WHITESPACE, 'Pe)'): RawTokenKind.RIGHT_PAREN,
    (RawTokenKind.WHITESPACE, 'Z'): RawTokenKind.WHITESPACE,
    (RawTokenKind.WHITESPACE, 'Cc'): RawTokenKind.WHITESPACE,
    (RawTokenKind.WHITESPACE, 'Po#'): RawTokenKind.COMMENT,
    (RawTokenKind.WHITESPACE, 'L'): RawTokenKind.ATOM,
    (RawTokenKind.WHITESPACE, 'Pd'): RawTokenKind.ATOM,
    (RawTokenKind.WHITESPACE, 'Pc'): RawTokenKind.ATOM,
    (RawTokenKind.WHITESPACE, 'N'): RawTokenKind.ATOM,
    (RawTokenKind.WHITESPACE, 'S'): RawTokenKind.OPERATOR,
    (RawTokenKind.WHITESPACE, 'Po'): RawTokenKind.OPERATOR,
    (RawTokenKind.LEFT_PAREN, 'Ps('): RawTokenKind.LEFT_PAREN,
    (RawTokenKind.LEFT_PAREN, 'Pe)'): RawTokenKind.RIGHT_PAREN,
    (RawTokenKind.LEFT_PAREN, 'Z'): RawTokenKind.WHITESPACE,
    (RawTokenKind.LEFT_PAREN, 'Cc'): RawTokenKind.WHITESPACE,
    (RawTokenKind.LEFT_PAREN, 'Po#'): RawTokenKind.COMMENT,
    (RawTokenKind.LEFT_PAREN, 'L'): RawTokenKind.ATOM,
    (RawTokenKind.LEFT_PAREN, 'Pd'): RawTokenKind.ATOM,
    (RawTokenKind.LEFT_PAREN, 'Pc'): RawTokenKind.ATOM,
    (RawTokenKind.LEFT_PAREN, 'N'): RawTokenKind.ATOM,
    (RawTokenKind.LEFT_PAREN, 'S'): RawTokenKind.OPERATOR,
    (RawTokenKind.LEFT_PAREN, 'Po'): RawTokenKind.OPERATOR,
    (RawTokenKind.RIGHT_PAREN, 'Ps('): RawTokenKind.LEFT_PAREN,
    (RawTokenKind.RIGHT_PAREN, 'Pe)'): RawTokenKind.RIGHT_PAREN,
    (RawTokenKind.RIGHT_PAREN, 'Z'): RawTokenKind.WHITESPACE,
    (RawTokenKind.RIGHT_PAREN, 'Cc'): RawTokenKind.WHITESPACE,
    (RawTokenKind.RIGHT_PAREN, 'Po#'): RawTokenKind.COMMENT,
    (RawTokenKind.RIGHT_PAREN, 'L'): RawTokenKind.ATOM,
    (RawTokenKind.RIGHT_PAREN, 'Pd'): RawTokenKind.ATOM,
    (RawTokenKind.RIGHT_PAREN, 'Pc'): RawTokenKind.ATOM,
    (RawTokenKind.RIGHT_PAREN, 'N'): RawTokenKind.ATOM,
    (RawTokenKind.RIGHT_PAREN, 'S'): RawTokenKind.OPERATOR,
    (RawTokenKind.RIGHT_PAREN, 'Po'): RawTokenKind.OPERATOR,
    (RawTokenKind.COMMENT, 'Cc'): RawTokenKind.WHITESPACE,
    (RawTokenKind.COMMENT, 'L'): RawTokenKind.COMMENT,
    (RawTokenKind.COMMENT, 'N'): RawTokenKind.COMMENT,
    (RawTokenKind.COMMENT, 'P'): RawTokenKind.COMMENT,
    (RawTokenKind.COMMENT, 'S'): RawTokenKind.COMMENT,
    (RawTokenKind.COMMENT, 'Z'): RawTokenKind.COMMENT,
    (RawTokenKind.ATOM, 'Ps('): RawTokenKind.LEFT_PAREN,
    (RawTokenKind.ATOM, 'Pe)'): RawTokenKind.RIGHT_PAREN,
    (RawTokenKind.ATOM, 'Z'): RawTokenKind.WHITESPACE,
    (RawTokenKind.ATOM, 'Cc'): RawTokenKind.WHITESPACE,
    (RawTokenKind.ATOM, 'Po#'): RawTokenKind.COMMENT,
    (RawTokenKind.ATOM, 'L'): RawTokenKind.ATOM,
    (RawTokenKind.ATOM, 'Pd'): RawTokenKind.ATOM,
    (RawTokenKind.ATOM, 'Pc'): RawTokenKind.ATOM,
    (RawTokenKind.ATOM, 'N'): RawTokenKind.ATOM,
    (RawTokenKind.ATOM, 'S'): RawTokenKind.OPERATOR,
    (RawTokenKind.ATOM, 'Po'): RawTokenKind.ATOM,
    (RawTokenKind.OPERATOR, 'Ps('): RawTokenKind.LEFT_PAREN,
    (RawTokenKind.OPERATOR, 'Pe)'): RawTokenKind.RIGHT_PAREN,
    (RawTokenKind.OPERATOR, 'Z'): RawTokenKind.WHITESPACE,
    (RawTokenKind.OPERATOR, 'Cc'): RawTokenKind.WHITESPACE,
    (RawTokenKind.OPERATOR, 'Po#'): RawTokenKind.COMMENT,
    (RawTokenKind.OPERATOR, 'L'): RawTokenKind.ATOM,
    (RawTokenKind.OPERATOR, 'Pd'): RawTokenKind.ATOM,
    (RawTokenKind.OPERATOR, 'Pc'): RawTokenKind.ATOM,
    (RawTokenKind.OPERATOR, 'N'): RawTokenKind.ATOM,
    (RawTokenKind.OPERATOR, 'S'): RawTokenKind.OPERATOR,
    (RawTokenKind.OPERATOR, 'Po'): RawTokenKind.OPERATOR,
}


def raw_tokenize(src: str) -> Iterable[RawToken]:
    """
    Split the string `src` into an iterable of `RawToken`s.

    The raw tokens are sufficient to completely reproduce the input stream and
    so may be useful for editing applications, most users will probably want
    `tokenize`.
    """
    # Raw token handling; there is a later semantic mapping stage which
    # annotates atoms for the special handling of keywords and numbers.
    # We treat tokenization as an explicit state machine.
    # State transitions emit the previous block along with the previous state.
    state, start = None, 0

    for index, character in enumerate(src):
        next_state = None
        major_category = unicodedata.category(character) + character

        for (from_state, category_match), to_state in STATE_MACHINE.items():
            if (
                from_state == state and
                major_category.startswith(category_match)
            ):
                next_state = to_state
                break

        if next_state is None:
            raise ParseError(
                "Unexpected '{0!r}'".format(character),
                (index, index + 1),
            )

        if next_state != state:
            if start != index:
                assert state is not None

                yield RawToken(
                    kind=state,
                    value=src[start:index],
                    location=(start, index),
                )
            start = index
            state = next_state

    if start != len(src):
        assert state is not None

        yield RawToken(
            kind=state,
            value=src[start:],
            location=(start, index + 1),
        )


def _digest_atom(raw_token: RawToken) -> Token:
    """Turn a raw `Atom` token into the more specific kinds."""

    # Is this a preposition
    try:
        return Token(
            kind=TokenKind.PREPOSITION,
            value=Preposition(raw_token.value),
            location=raw_token.location,
        )
    except ValueError:
        pass

    # Is this an integer value?
    try:
        return Token(
            kind=TokenKind.NUMBER,
            value=int(raw_token.value),
            location=raw_token.location,
        )
    except ValueError:
        pass

    # A float?
    try:
        return Token(
            kind=TokenKind.NUMBER,
            value=float(raw_token.value),
            location=raw_token.location,
        )
    except ValueError:
        pass

    # A duration?
    duration_match = RE_DURATION.match(raw_token.value)
    if duration_match is not None:
        total_length = (
            int(duration_match.group(1) or '0') * 24 * 60 * 60 +
            int(duration_match.group(2) or '0') * 60 * 60 +
            int(duration_match.group(3) or '0') * 60 +
            int(duration_match.group(4) or '0')
        )
        return Token(
            kind=TokenKind.DURATION,
            value=total_length,
            location=raw_token.location,
        )

    # A literal?
    try:
        kind, value = LITERALS[raw_token.value]
        return Token(
            kind=kind,
            value=value,
            location=raw_token.location,
        )
    except KeyError:
        pass

    # Pass through, split by dots
    return Token(
        kind=TokenKind.ATOM,
        value=raw_token.value.split('.'),
        location=raw_token.location,
    )


def tokenize(src: str) -> Iterable[Token]:
    """Split the string `src` into an iterable of `Token`s."""
    for raw_token in raw_tokenize(src):
        if (
            raw_token.kind in (RawTokenKind.COMMENT, RawTokenKind.WHITESPACE)
        ):
            continue

        if raw_token.kind == RawTokenKind.ATOM:
            yield _digest_atom(raw_token)
        else:
            yield Token(
                kind=RAW_TOKEN_KIND_TO_TOKEN_KIND[raw_token.kind],
                value=raw_token.value,
                location=raw_token.location,
            )



"""
Versioning utils.
"""

__all__ = ('get_version')

import os
import re
import logging
import os.path
import subprocess
from os.path import dirname

import pkg_resources

version_re = re.compile('^Version: (.+)$', re.M)

logger = logging.getLogger(__file__)


def find_git_root(test):
    prev, test = None, os.path.abspath(test)
    while prev != test:
        if os.path.isdir(os.path.join(test, '.git')):
            return test
        prev, test = test, os.path.abspath(os.path.join(test, os.pardir))
    return None


def get_version():
    """
    Gets the current version number.

    If in a git repository, it is the current git tag.
    Otherwise it is the one contained in the PKG-INFO file.
    To use this script, simply import it in your setup.py file
    and use the results of get_version() as your package version:
        from version import *
        setup(
            ...
            version=get_version(),
            ...
        )
    """
    git_root = find_git_root(dirname(__file__))

    if git_root is not None:
        # Get the version using "git describe".
        cmd = 'git describe --tags --match [0-9]*'.split()
        try:
            version = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            logger.exception('Unable to get version number from git tags')
            exit(1)

        # PEP 386 compatibility
        if '-' in version:
            version = '.post'.join(version.split('-')[:2])

        # Don't declare a version "dirty" merely because a time stamp has
        # changed. If it is dirty, append a ".dev1" suffix to indicate a
        # development revision after the release.
        with open(os.devnull, 'w') as fd_devnull:
            subprocess.call(
                ['git', 'status'],
                stdout=fd_devnull,
                stderr=fd_devnull,
            )

        cmd = 'git diff-index --name-only HEAD'.split()
        try:
            dirty = subprocess.check_output(cmd).decode().strip()
        except subprocess.CalledProcessError:
            logger.exception('Unable to get git index status')
            exit(1)

        if dirty != '':
            version += '.dev1'

        return version

    else:
        try:
            return pkg_resources.working_set.by_key['routemaster'].version
        except KeyError:
            return '0.0.0-unreleased'



"""Exit condition program evaluator."""

from routemaster.exit_conditions.operations import Operation


def _evaluate_to_bool(stack, lookup, property_handler):
    top_of_stack = stack.pop()
    stack.append(bool(top_of_stack))


def _evaluate_not(stack, lookup, property_handler):
    top_of_stack = stack.pop()
    stack.append(not top_of_stack)


def _evaluate_and(stack, lookup, property_handler):
    rhs = stack.pop()
    lhs = stack.pop()
    stack.append(lhs and rhs)


def _evaluate_or(stack, lookup, property_handler):
    rhs = stack.pop()
    lhs = stack.pop()
    stack.append(lhs or rhs)


def _evaluate_literal(stack, lookup, property_handler, value):
    stack.append(value)


def _evaluate_lookup(stack, lookup, property_handler, key):
    stack.append(lookup(key))


def _evaluate_eq(stack, lookup, property_handler):
    rhs = stack.pop()
    lhs = stack.pop()
    stack.append(lhs == rhs)


def _evaluate_lt(stack, lookup, property_handler):
    rhs = stack.pop()
    lhs = stack.pop()
    stack.append(lhs < rhs)


def _evaluate_gt(stack, lookup, property_handler):
    rhs = stack.pop()
    lhs = stack.pop()
    stack.append(lhs > rhs)


def _evaluate_property(
    stack,
    lookup,
    property_handler,
    property_name,
    prepositions,
):
    prepositional_arguments = {}
    for preposition in reversed(prepositions):
        prepositional_arguments[preposition.value] = stack.pop()
    subject = stack.pop()
    stack.append(
        property_handler(property_name, subject, **prepositional_arguments),
    )


EVALUATORS = {
    Operation.TO_BOOL: _evaluate_to_bool,
    Operation.AND: _evaluate_and,
    Operation.OR: _evaluate_or,
    Operation.NOT: _evaluate_not,
    Operation.PROPERTY: _evaluate_property,
    Operation.GT: _evaluate_gt,
    Operation.LT: _evaluate_lt,
    Operation.EQ: _evaluate_eq,
    Operation.LITERAL: _evaluate_literal,
    Operation.LOOKUP: _evaluate_lookup,
}


def evaluate(instructions, lookup, property_handler):
    """
    Run the instructions given in `instructions`.

    Returns the single result.
    """
    stack = []
    for instruction, *args in instructions:
        EVALUATORS[instruction](stack, lookup, property_handler, *args)
    return stack.pop()



"""Logger for multiple backends."""

import functools
import contextlib
from typing import List

from routemaster.logging.base import BaseLogger


class SplitLogger(BaseLogger):
    """Proxies logging calls to all loggers in a list."""

    def __init__(self, *args, loggers: List[BaseLogger]) -> None:
        super().__init__(*args)

        self.loggers = loggers

        for fn in (
            'init_flask',

            'debug',
            'info',
            'warning',
            'error',
            'critical',
            'exception',

            'webhook_response',
            'feed_response',
            'process_request_started',
            'process_request_finished',
        ):
            setattr(self, fn, functools.partial(self._log_all, fn))

        for fn in (
            'process_cron',
            'process_webhook',
            'process_feed',
        ):
            setattr(self, fn, functools.partial(self._log_all_ctx, fn))

    def _log_all(self, name, *args, **kwargs):
        for logger in self.loggers:
            getattr(logger, name)(*args, **kwargs)

    @contextlib.contextmanager
    def _log_all_ctx(self, name, *args, **kwargs):
        with contextlib.ExitStack() as stack:
            for logger in self.loggers:
                logger_ctx = getattr(logger, name)
                stack.enter_context(logger_ctx(*args, **kwargs))
            yield



"""The core of the state machine logic."""

from typing import List, Callable, Iterable, Optional

from typing_extensions import Protocol

from routemaster.db import Label, History
from routemaster.app import App
from routemaster.utils import dict_merge, suppress_exceptions
from routemaster.config import Gate, State, StateMachine
from routemaster.state_machine.gates import process_gate
from routemaster.state_machine.types import LabelRef, Metadata
from routemaster.state_machine.utils import (
    lock_label,
    get_current_state,
    get_state_machine,
)
from routemaster.state_machine.utils import \
    get_label_metadata as get_label_metadata_internal
from routemaster.state_machine.utils import (
    needs_gate_evaluation_for_metadata_change,
)
from routemaster.state_machine.exceptions import (
    DeletedLabel,
    UnknownLabel,
    LabelAlreadyExists,
)
from routemaster.state_machine.transitions import process_transitions

# Signature of a function to gather the labels to be operated upon when
# processing a cron task. This will be called in a different transaction to
# where we iterate over the results, so to prevent confusion or the possible
# introduction of errors, we require all the data up-front.
LabelProvider = Callable[[App, StateMachine, State], List[str]]


def list_labels(app: App, state_machine: StateMachine) -> Iterable[LabelRef]:
    """
    Returns a sorted iterable of labels associated with a state machine.

    Labels are returned ordered alphabetically by name.
    """
    for (name,) in app.session.query(Label.name).filter_by(
        state_machine=state_machine.name,
        deleted=False,
    ):
        yield LabelRef(name=name, state_machine=state_machine.name)


def get_label_state(app: App, label: LabelRef) -> Optional[State]:
    """Finds the current state of a label."""
    state_machine = get_state_machine(app, label)
    return get_current_state(app, label, state_machine)


def get_label_metadata(app: App, label: LabelRef) -> Metadata:
    """Returns the metadata associated with a label."""
    state_machine = get_state_machine(app, label)

    row = get_label_metadata_internal(app, label, state_machine)

    if row is None:
        raise UnknownLabel(label)

    metadata, deleted = row

    if deleted:
        raise DeletedLabel(label)

    return metadata


def create_label(app: App, label: LabelRef, metadata: Metadata) -> Metadata:
    """Creates a label and starts it in a state machine."""
    state_machine = get_state_machine(app, label)

    if app.session.query(
        app.session.query(Label).filter_by(
            name=label.name,
            state_machine=label.state_machine,
        ).exists(),
    ).scalar():
        raise LabelAlreadyExists(label)

    app.session.add(
        Label(
            name=label.name,
            state_machine=state_machine.name,
            metadata=metadata,
            history=[
                History(
                    old_state=None,
                    new_state=state_machine.states[0].name,
                ),
            ],
        ),
    )

    process_transitions(app, label)

    return metadata


def update_metadata_for_label(
    app: App,
    label: LabelRef,
    update: Metadata,
) -> Metadata:
    """
    Updates the metadata for a label.

    Moves the label through the state machine as appropriate.
    """
    state_machine = get_state_machine(app, label)
    needs_gate_evaluation = False

    row = lock_label(app, label)

    existing_metadata, deleted = row.metadata, row.deleted
    if deleted:
        raise DeletedLabel(label)

    needs_gate_evaluation, current_state = \
        needs_gate_evaluation_for_metadata_change(
            app,
            state_machine,
            label,
            update,
        )

    new_metadata = dict_merge(existing_metadata, update)

    row.metadata = new_metadata
    row.metadata_triggers_processed = not needs_gate_evaluation

    # Try to move the label forward, but this is not a hard requirement as
    # the cron will come back around to progress the label later.
    if needs_gate_evaluation:
        try:
            _process_transitions_for_metadata_update(
                app,
                label,
                state_machine,
                current_state,
            )
        except Exception:
            # This is allowed to fail here. We have successfully saved the new
            # metadata, and it has a metadata_triggers_processed=False flag so
            # will be picked up again for processing later.
            app.logger.exception(
                f"Failed to progress label {label!r} after metadata update.",
            )

    return new_metadata


def _process_transitions_for_metadata_update(
    app: App,
    label: LabelRef,
    state_machine: StateMachine,
    state_pending_update: State,
):
    with app.session.begin_nested():
        lock_label(app, label)
        current_state = get_current_state(app, label, state_machine)

        if state_pending_update != current_state:
            # We have raced with another update, and are no longer in
            # the state for which we needed an update, so we should
            # stop.
            return

        if not isinstance(current_state, Gate):  # pragma: no branch
            # Cannot be hit because of the semantics of
            # `needs_gate_evaluation_for_metadata_change`. Here to
            # appease mypy.
            raise RuntimeError(  # pragma: no cover
                "Label not in a gate",
            )

        could_progress = process_gate(
            app=app,
            state=current_state,
            state_machine=state_machine,
            label=label,
        )

    if could_progress:
        process_transitions(app, label)


def delete_label(app: App, label: LabelRef) -> None:
    """
    Deletes the metadata for a label and marks the label as deleted.

    The history for the label is not changed (in order to allow post-hoc
    analysis of the path the label took through the state machine).
    """
    state_machine = get_state_machine(app, label)  # Raises UnknownStateMachine

    try:
        row = lock_label(app, label)
    except UnknownLabel:
        return

    if row is None or row.deleted:
        return

    # Record the label as having been deleted and remove its metadata
    row.metadata = {}
    row.deleted = True

    # Add a history entry for the deletion
    current_state = get_current_state(app, label, state_machine)

    if current_state is None:
        raise AssertionError(f"Active label {label} has no current state!")

    row.history.append(History(
        old_state=current_state.name,
        new_state=None,
    ))


class LabelStateProcessor(Protocol):
    """Type signature for the label state processor callable."""
    def __call__(
        self,
        *,
        app: App,
        state: State,
        state_machine: StateMachine,
        label: LabelRef,
    ) -> bool:
        """Type signature for the label state processor callable."""
        ...

    def __name__(self) -> str:
        """Name of the callable."""
        ...


def process_cron(
    process: LabelStateProcessor,
    get_labels: Callable[[StateMachine, State], Iterable[str]],
    app: App,
    state_machine: StateMachine,
    state: State,
):
    """
    Cron event entrypoint.
    """
    with app.new_session():
        relevant_labels = get_labels(state_machine, state)

    for label_name in relevant_labels:
        with suppress_exceptions(app.logger):
            label = LabelRef(name=label_name, state_machine=state_machine.name)
            could_progress = False

            with app.new_session():
                lock_label(app, label)
                current_state = get_current_state(app, label, state_machine)

                if current_state != state:
                    continue

                could_progress = process(
                    app=app,
                    state=state,
                    state_machine=state_machine,
                    label=label,
                )

            with app.new_session():
                if could_progress:
                    process_transitions(app, label)



"""Webhook invocation."""

import enum
from typing import Any, Dict, Callable, Iterable

import requests

from routemaster.config import Webhook, StateMachine


@enum.unique
class WebhookResult(enum.Enum):
    """Possible results from invoking a webhook."""
    SUCCESS = 'success'
    RETRY = 'retry'
    FAIL = 'fail'


ResponseLogger = Callable[[requests.Response], None]
WebhookRunner = Callable[[str, str, bytes, str, ResponseLogger], WebhookResult]


class RequestsWebhookRunner(object):
    """
    Webhook runner which uses `requests` to actually hit webhooks.

    Optionally takes a list of webhook configs to modify how requests are made.
    """

    def __init__(self, webhook_configs: Iterable[Webhook] = ()) -> None:
        # Use a session so that we can take advantage of connection pooling in
        # `urllib3`.
        self.session = requests.Session()
        self.webhook_configs = webhook_configs

    def __call__(
        self,
        url: str,
        content_type: str,
        data: bytes,
        idempotency_token: str,
        log_response: ResponseLogger = lambda x: None,
    ) -> WebhookResult:
        """Run a POST on the given webhook."""
        headers = {
            'Content-Type': content_type,
            'X-Idempotency-Token': idempotency_token,
        }
        headers.update(self._headers_for_url(url))

        try:
            result = self.session.post(
                url,
                data=data,
                headers=headers,
                timeout=10,
            )
            log_response(result)
        except requests.exceptions.RequestException:
            return WebhookResult.RETRY

        if result.status_code == 410:
            return WebhookResult.FAIL
        elif 200 <= result.status_code < 300:
            return WebhookResult.SUCCESS
        else:
            return WebhookResult.RETRY

    def _headers_for_url(self, url: str) -> Dict[str, Any]:
        headers = {}
        for config in self.webhook_configs:
            if config.match.search(url):
                headers.update(config.headers)
        return headers


def webhook_runner_for_state_machine(
    state_machine: StateMachine,
) -> WebhookRunner:
    """
    Create the webhook runner for a given state machine.

    Applies any state machine configuration to the runner.
    """
    return RequestsWebhookRunner(state_machine.webhooks)



"""Helpers for working with time."""

import datetime


def time_appears_in_range(
    when: datetime.time,
    start: datetime.datetime,
    end: datetime.datetime,
) -> bool:
    """
    Determine whether the given time appears within the given datetime range.

    Note: the comparison does not include the start instant, but does include
    the end instant.
    """
    if start >= end:
        raise ValueError(
            f"Must be passed a valid range to check (got {start} until {end})",
        )

    if (end - start) >= datetime.timedelta(days=1):
        return True

    date = start.date() if when > start.time() else end.date()

    instant = datetime.datetime.combine(date, when)

    return start < instant <= end



"""Analysis of compiled programs."""


from routemaster.exit_conditions.operations import Operation


def find_accessed_keys(instructions):
    """Yield each key accessed under the program."""
    for instruction, *args in instructions:
        if instruction == Operation.LOOKUP:
            yield args[0]



"""Operation definitions for the condition stack machine."""

import enum


@enum.unique
class Operation(enum.Enum):
    """Classes of instructions for the exit condition machine."""

    # Pop `value` from the stack, interpret it as a boolean, and push
    # that to the stack.
    TO_BOOL = 'to_bool'

    # Pop boolean `value` from the stack, negate it, and push the negated
    # form to the stack.
    NOT = 'not'

    # Pop boolean `rhs` and `lhs` from the stack, AND together, and push
    # that to the stack.
    AND = 'and'

    # Pop boolean `rhs` and `lhs` from the stack, OR together, and push
    # that to the stack.
    OR = 'or'

    # Push literal value `argument` to the stack.
    LITERAL = 'literal'

    # Look up symbol `argument` and push it to the stack, or None if not found.
    LOOKUP = 'lookup'

    # Pop boolean `rhs` and `lhs` from the stack, compare equality, and push
    # `true` if equal and `false` if not.
    EQ = 'eq'

    # Pop boolean `rhs` and `lhs` from the stack, compare ordering, and push
    # `true` if `rhs` is less than `lhs` and `false` if not.
    LT = 'lt'

    # Pop boolean `rhs` and `lhs` from the stack, compare ordering, and push
    # `true` if `rhs` is greater than `lhs` and `false` if not.
    GT = 'gt'

    # Pop value `value` from the stack, establish if it matches condition
    # `argument` and push `true` if so, `false` if not.
    PROPERTY = 'property'



"""Public Database interface."""

from routemaster.db.model import Label, History, metadata
from routemaster.db.initialisation import initialise_db

__all__ = (
    'Label',
    'History',
    'metadata',
    'initialise_db',
)



"""Plugin loading and configuration."""
import importlib

from routemaster.config import Config, LoggingPluginConfig
from routemaster.logging.base import BaseLogger


class PluginConfigurationException(Exception):
    """Raised to signal an invalid plugin that was loaded."""


def register_loggers(config: Config):
    """
    Iterate through all plugins in the config file and instatiate them.
    """
    return [_import_logger(config, x) for x in config.logging_plugins]


def _import_logger(
    config: Config,
    logger_config: LoggingPluginConfig,
) -> BaseLogger:
    dotted_path = logger_config.dotted_path

    try:
        module_path, klass_name = dotted_path.rsplit(':', 2)
    except ValueError:
        raise PluginConfigurationException(
            f"{dotted_path} must be in the form <module-path>:<class-name>",
        )

    try:
        module = importlib.import_module(module_path)
    except ImportError:
        raise PluginConfigurationException(
            f"{module_path} does not exist on the PYTHONPATH",
        )

    try:
        klass = getattr(module, klass_name)
    except AttributeError:
        raise PluginConfigurationException(
            f"{klass_name} does not exist in module {module_path}",
        )

    if not callable(klass):
        raise PluginConfigurationException(
            f"{dotted_path} must be callable",
        )

    try:
        logger = klass(config, **logger_config.kwargs)
    except TypeError:
        raise PluginConfigurationException(
            f"Could not instantiate logger, {klass_name} must take a config "
            f"argument and any kwargs specified in the plugin configuration.",
        )

    if not isinstance(logger, BaseLogger):
        raise PluginConfigurationException(
            f"{dotted_path} must inherit from routemaster.logging.BaseLogger",
        )

    return logger



"""Parser and compiler for exit conditions."""

from routemaster.exit_conditions.tokenizer import TokenKind, tokenize
from routemaster.exit_conditions.exceptions import ParseError
from routemaster.exit_conditions.operations import Operation

OPERATOR_DID_YOU_MEAN = {
    '==': '=',
    '!=': '/=',
    '<>': '/=',
}


class _TokenSource(object):
    """A source of tokens in a form which is convenient for parsing."""

    def __init__(self, iterable):
        self.iterable = iterable
        self.previous_location = None
        self.head = object()
        self._advance()

    def _advance(self):
        assert self.head is not None

        try:
            self.previous_location = getattr(self.head, 'location', None)
            self.head = next(self.iterable)
        except StopIteration:
            self.head = None

    def try_eat_next(self, *kinds):
        if not self.match_next(*kinds):
            return False

        self._advance()
        return True

    def match_next(self, *kinds):
        if self.head is None:
            return False

        if self.head.kind in kinds:
            return True
        return False

    def eat_next(self, *kinds):
        if self.head is None:
            if self.previous_location is None:
                # Empty program
                raise ParseError(
                    "Surprisingly, program was empty",
                    location=(0, 1),
                )

            end_of_last_location = self.previous_location[-1]
            raise ParseError(
                "Unexpected EOF, expected {kind}".format(
                    kind=', '.join(x.value for x in kinds),
                ),
                location=(
                    end_of_last_location,
                    end_of_last_location + 1,
                ),
            )

        if self.head.kind in kinds:
            old_head = self.head
            self._advance()
            return old_head
        raise ParseError("Expected {kind}, got {actual_kind}".format(
            kind=', '.join(x.value for x in kinds),
            actual_kind=self.head.kind.value,
        ), self.head.location)


def _parse_and_expr(source):
    already_bool_converted = False

    yield from _parse_or_expr(source)

    while source.try_eat_next(TokenKind.AND):
        if not already_bool_converted:
            already_bool_converted = True
            yield (Operation.TO_BOOL,)
        yield from _parse_or_expr(source)
        yield (Operation.TO_BOOL,)
        yield (Operation.AND,)


def _parse_or_expr(source):
    already_bool_converted = False

    yield from _parse_base_expr(source)

    while source.try_eat_next(TokenKind.OR):
        if not already_bool_converted:
            already_bool_converted = True
            yield (Operation.TO_BOOL,)
        yield from _parse_base_expr(source)
        yield (Operation.TO_BOOL,)
        yield (Operation.OR,)


def _parse_base_expr(source):
    negated = False
    while source.try_eat_next(TokenKind.NOT):
        negated = not negated

    yield from _parse_value(source)
    if source.try_eat_next(TokenKind.COPULA):
        # `is` or `has` expression
        if source.try_eat_next(TokenKind.NOT):
            negated = not negated

        try:
            adjective = tuple(source.eat_next(TokenKind.ATOM).value)
        except ParseError:
            # Must be a prepositional phrase
            adjective = ()

        prepositions = []
        while source.match_next(TokenKind.PREPOSITION):
            prepositions.append(source.head.value)
            source.eat_next(TokenKind.PREPOSITION)
            yield from _parse_value(source)

        if not adjective and not prepositions:
            if source.head is not None:
                raise ParseError(
                    "Expected an adjective or preposition",
                    source.head.location,
                )
            else:
                raise ParseError(
                    "Expected an adjective or preposition afterwards, "
                    "but got the EOF",
                    source.previous_location,
                )

        yield Operation.PROPERTY, tuple(adjective), tuple(prepositions)

    elif source.match_next(TokenKind.OPERATOR):
        try:
            operator, is_negative = {
                '=': (Operation.EQ, False),
                '/=': (Operation.EQ, True),
                '<': (Operation.LT, False),
                '>': (Operation.GT, False),
                '<=': (Operation.GT, True),
                '>=': (Operation.LT, True),
            }[source.head.value]
        except KeyError:
            try:
                suggested_replacement = \
                    OPERATOR_DID_YOU_MEAN[source.head.value]
                raise ParseError(
                    "Unknown operator {operator} "
                    "(did you mean {suggestion}?)".format(
                        operator=source.head.value,
                        suggestion=suggested_replacement,
                    ),
                    location=source.head.location,
                ) from None
            except KeyError:
                raise ParseError(
                    "Unknown operator {operator}".format(
                        operator=source.head.value,
                    ),
                    location=source.head.location,
                ) from None

        if is_negative:
            negated = not negated
        source.eat_next(TokenKind.OPERATOR)

        yield from _parse_value(source)
        yield (operator,)

    if negated:
        yield (Operation.TO_BOOL,)
        yield (Operation.NOT,)


def _parse_value(source):
    # Immediate special-case: parentheticals
    if source.try_eat_next(TokenKind.LEFT_PAREN):
        yield from _parse_and_expr(source)
        source.eat_next(TokenKind.RIGHT_PAREN)
        return

    # Atomic lookup
    try:
        atom = source.eat_next(TokenKind.ATOM)
        yield Operation.LOOKUP, tuple(atom.value)
        return
    except ParseError:
        pass

    # Literals
    try:
        literal = source.eat_next(TokenKind.LITERAL)
        yield Operation.LITERAL, literal.value
        return
    except ParseError:
        pass

    # Durations
    try:
        duration = source.eat_next(TokenKind.DURATION)
        yield Operation.LITERAL, duration.value
        return
    except ParseError:
        pass

    # Numbers
    try:
        number = source.eat_next(TokenKind.NUMBER)
        yield Operation.LITERAL, number.value
        return
    except ParseError:
        pass

    # No match
    if source.head is not None:
        raise ParseError(
            "Expected a value, got {kind}".format(
                kind=source.head.kind.value,
            ),
            location=source.head.location,
        )
    else:
        if source.previous_location is not None:
            raise ParseError(
                "Expected a value, but the EOF was reached",
                location=source.previous_location,
            )
        else:
            raise ParseError(
                "Expected a value but this program is empty",
                location=(0, 1),
            )


def _parse_tokens(token_stream):
    source = _TokenSource(token_stream)
    yield from _parse_and_expr(source)
    # Always end in a final TO_BOOL
    yield (Operation.TO_BOOL,)

    # Verify that this is actually the entire stream
    if source.head is not None:
        raise ParseError(
            f"Unexpected token after end of program: {source.head.kind.value}",
            location=source.head.location,
        )


def parse(source):
    """Compile from arbitrary source to a sequence of program instructions."""
    return tuple(_parse_tokens(tokenize(source)))



"""Loading of application configuration."""

from routemaster.config.model import (
    Gate,
    State,
    Action,
    Config,
    Trigger,
    Webhook,
    FeedConfig,
    NextStates,
    NoNextStates,
    StateMachine,
    DatabaseConfig,
    OnEntryTrigger,
    IntervalTrigger,
    MetadataTrigger,
    ConstantNextState,
    ContextNextStates,
    SystemTimeTrigger,
    LoggingPluginConfig,
    TimezoneAwareTrigger,
    ContextNextStatesOption,
    MetadataTimezoneAwareTrigger,
)
from routemaster.config.loader import load_config, load_database_config
from routemaster.config.exceptions import ConfigError

__all__ = (
    'load_config',
    'load_database_config',
    'Gate',
    'State',
    'Action',
    'Config',
    'Trigger',
    'Webhook',
    'FeedConfig',
    'NextStates',
    'ConfigError',
    'NoNextStates',
    'StateMachine',
    'DatabaseConfig',
    'OnEntryTrigger',
    'IntervalTrigger',
    'MetadataTrigger',
    'ConstantNextState',
    'ContextNextStates',
    'SystemTimeTrigger',
    'LoggingPluginConfig',
    'TimezoneAwareTrigger',
    'ContextNextStatesOption',
    'MetadataTimezoneAwareTrigger',
)



"""State machine exceptions."""


class UnknownLabel(ValueError):
    """Represents a label unknown in the given state machine."""
    deleted = False


class DeletedLabel(UnknownLabel):
    """Represents a label deleted in the given state machine."""
    deleted = True


class UnknownStateMachine(ValueError):
    """Represents a state machine not in the system."""


class LabelAlreadyExists(ValueError):
    """Thrown when a label already exists in the state machine."""



"""Human-readable ParseError handling."""


def _find_line_containing(source, index):
    """Find (line number, line, offset) triple for an index into a string."""
    lines = source.splitlines()

    if not lines:
        # Special case: empty program
        return 1, '', 0

    this_line_start = 0
    for zero_index_line_number, line in enumerate(lines):
        next_line_start = this_line_start + len(line) + 1
        if next_line_start > index:
            return zero_index_line_number + 1, line, index - this_line_start
        this_line_start = next_line_start

    # Must be at the end of file
    raise AssertionError("index >> len(source)")


def format_parse_error_message(*, source, error):
    """Format a parse error on some source for nicer display."""
    error_line_number, error_line, error_offset = _find_line_containing(
        source,
        error.location[0],
    )

    error_length = error.location[1] - error.location[0]

    error_character = "~" if error_length > 1 else "^"

    message_lines = [
        f"Error on line {error_line_number}: {error.message}",
        error_line,
        " " * error_offset + error_character * error_length,
    ]

    return "\n".join(message_lines)
